import threading


class Canceller:
    def __init__(self):
        self._cancelled = False
        self._lock = threading.Lock()

    def cancel(self):
        with self._lock:
            self._cancelled = True

    def is_cancelled(self) -> bool:
        with self._lock:
            return self._cancelled



import asyncio
from typing import AsyncIterator, Iterator, TypeVar

T = TypeVar("T")


def async_iterable_to_sync_iterable(iterator: AsyncIterator[T]) -> Iterator[T]:
    with asyncio.Runner() as runner:
        try:
            while True:
                result = runner.run(anext(iterator))  # type: ignore
                yield result
        except StopAsyncIteration:
            pass



import os
import shutil
import subprocess
from io import TextIOBase
from typing import Literal

from handbrake.canceller import Canceller
from handbrake.models.preset import Preset, PresetGroup, PresetInfo
from handbrake.models.progress import Progress
from handbrake.models.title import TitleSet
from handbrake.models.version import Version
from handbrake.opts import ConvertOpts, generate_convert_args, generate_scan_args
from handbrake.progresshandler import ProgressHandler
from handbrake.runner import (
    ConvertCommandRunner,
    PresetCommandRunner,
    ScanCommandRunner,
    VersionCommandRunner,
)


class HandBrake:
    def __init__(self, executable: str | None = None):
        """Initialise the HandBrake wrapper

        :param executable: path to the HandBrakeCLI executable to
        use. If not provided, an attempt to locate the executable is
        made by checking the following, setting the executable to the
        first valid option: the value of the environment variable
        HANDBRAKECLI, examining the PATH variable for a HandBrakeCLI
        executable and examining the PATH for a handbrakecli
        executable

        """
        if executable is not None:
            self.executable = executable
        elif e := os.getenv("HANDBRAKECLI"):
            self.executable = e
        elif w := shutil.which("HandBrakeCLI"):
            self.executable = w
        elif w := shutil.which("handbrakecli"):
            self.executable = w
        else:
            raise FileNotFoundError("could not find HandBrakeCLI")

    def version(self) -> Version:
        """Returns the version of HandBrakeCLI

        :returns: an object holding the handbrake version
        """
        version: Version | None = None
        runner = VersionCommandRunner()
        args = ["--json", "--version"]
        for obj in runner.process(self.executable, *args):
            if isinstance(obj, Version):
                version = obj

        if version is None:
            raise RuntimeError("version not found")
        return version

    async def version_async(self, cancel: Canceller | None = None) -> Version:
        """Asynchronously returns the version of HandBrakeCLI

        :param cancel: a parameter that allows early termination of the command
        :returns: an object holding the handbrake version
        """
        version: Version | None = None
        runner = VersionCommandRunner()
        args = ["--json", "--version"]
        async for obj in runner.aprocess(self.executable, *args, cancel=cancel):
            if isinstance(obj, Version):
                version = obj

        if version is None:
            raise RuntimeError("version not found")
        return version

    def convert_title(
        self,
        input: str | os.PathLike,
        output: str | os.PathLike,
        title: int | Literal["main"],
        opts: ConvertOpts | None = None,
        progress_handler: ProgressHandler | None = None,
    ):
        """Convert a title from the input source

        :param input: the input source
        :param output: the path to write the converted file to
        :param title: the title to convert, either by integer index or
        'main' to select the main title
        :param opts: conversion options
        :param progress_handler: a callback function to handle progress updates
        """
        args = generate_convert_args(input, output, title, opts)
        runner = ConvertCommandRunner()
        for obj in runner.process(self.executable, *args):
            if isinstance(obj, Progress):
                if progress_handler is not None:
                    progress_handler(obj)

    async def convert_title_async(
        self,
        input: str | os.PathLike,
        output: str | os.PathLike,
        title: int | Literal["main"],
        opts: ConvertOpts | None = None,
        progress_handler: ProgressHandler | None = None,
        cancel: Canceller | None = None,
    ):
        """Asynchronously convert a title from the input source

        :param input: the input source
        :param output: the path to write the converted file to
        :param title: the title to convert, either by integer index or
        'main' to select the main title
        :param opts: conversion options
        :param progress_handler: a callback function to handle progress updates
        :param cancel: a parameter that allows early termination of the command
        """
        args = generate_convert_args(input, output, title, opts)
        runner = ConvertCommandRunner()
        async for obj in runner.aprocess(self.executable, *args, cancel=cancel):
            if isinstance(obj, Progress):
                if progress_handler is not None:
                    progress_handler(obj)

    def scan_titles(
        self,
        input: str | os.PathLike,
        title: int | Literal["main", "all"],
        progress_handler: ProgressHandler | None = None,
    ) -> TitleSet:
        """Scans the selected title(s) and returns their details

        :param input: the input source
        :param title: the title(s) to scan, either by integer index,
        'main' to select the main title or 'all' to select all title
        :param progress_handler: a callback function to handle progress updates
        :return: a `TitleSet` containing the selected title
        """

        args = generate_scan_args(input, title)
        title_set: TitleSet | None = None
        runner = ScanCommandRunner()
        for obj in runner.process(self.executable, *args):
            if isinstance(obj, Progress):
                if progress_handler is not None:
                    progress_handler(obj)
            elif isinstance(obj, TitleSet):
                title_set = obj

        # check output
        if title_set is None:
            raise RuntimeError("no titles found")
        if title != "all" and len(title_set.title_list) == 0:
            raise RuntimeError("title does not contain specified title")
        return title_set

    async def scan_titles_async(
        self,
        input: str | os.PathLike,
        title: int | Literal["main", "all"],
        progress_handler: ProgressHandler | None = None,
        cancel: Canceller | None = None,
    ) -> TitleSet:
        """Asynchronously scans the selected title(s) and returns their details

        :param input: the input source
        :param title: the title(s) to scan, either by integer index,
        'main' to select the main title or 'all' to select all title
        :param progress_handler: a callback function to handle progress updates
        :param cancel: A parameter to allow early termination of the command
        :return: a `TitleSet` containing the selected title
        """

        args = generate_scan_args(input, title)
        title_set: TitleSet | None = None
        runner = ScanCommandRunner()
        async for obj in runner.aprocess(self.executable, *args, cancel=cancel):
            if isinstance(obj, Progress):
                if progress_handler is not None:
                    progress_handler(obj)
            elif isinstance(obj, TitleSet):
                title_set = obj

        # check output
        if title_set is None:
            raise RuntimeError("no titles found")
        if title != "all" and len(title_set.title_list) == 0:
            raise RuntimeError("title does not contain specified title")
        return title_set

    def get_preset(self, name: str) -> Preset:
        """Get the builtin preset with the given name

        :param name: the name of the preset to select
        :returns: a `Preset` object containing the selected preset
        """
        preset_list: Preset | None = None
        runner = PresetCommandRunner()
        args = [
            "--json",
            "-Z",
            name,
            "--preset-export",
            name,
        ]
        for obj in runner.process(self.executable, *args):
            if isinstance(obj, Preset):
                preset_list = obj

        if preset_list is None:
            raise RuntimeError("no preset list found")
        return preset_list

    def list_presets(self) -> list[PresetGroup]:
        """List all builtin presets

        :returns: a list of preset groups
        """
        res: list[PresetGroup] = []
        curgroup = PresetGroup(name="", presets=[])
        curpreset = PresetInfo(name="", description="")
        cmd = [self.executable, "-z"]
        proc = subprocess.run(cmd, capture_output=True, check=True)

        # the output is of the format
        # groupA/
        #     preset 1 name
        #         the description of the preset
        #     preset 2 name
        #         the second description, it can span
        #         multiple lines
        # groupB/
        #     ...
        for line in proc.stderr.decode().splitlines():
            if line.endswith("/"):
                curgroup = PresetGroup(name=line[:-1], presets=[])
                res.append(curgroup)
            elif line.startswith("        "):
                if curpreset.description == "":
                    curpreset.description = line.strip()
                else:
                    curpreset.description += " " + line.strip()
            elif line.startswith("    "):
                curpreset = PresetInfo(name=line.strip(), description="")
                curgroup.presets.append(curpreset)
        return res

    def load_preset_from_file(self, file: str | os.PathLike | TextIOBase) -> Preset:
        """Load a handbrake preset export into a `Preset` object

        :param file: either a filepath or a file-like object to read the preset from
        :returns: a `Preset` object from the data in the given file
        """
        if isinstance(file, TextIOBase):
            return Preset.model_validate_json(file.read())
        else:
            with open(file) as f:
                return Preset.model_validate_json(f.read())

    def save_preset_to_file(self, file: str | os.PathLike | TextIOBase, preset: Preset):
        """Save a handbrake preset to a file

        :param file: either a filepath or a file-like object to write the preset to
        """
        if isinstance(file, TextIOBase):
            file.write(preset.model_dump_json(by_alias=True))
        else:
            with open(file, "w") as f:
                f.write(preset.model_dump_json(by_alias=True))



import asyncio
import asyncio.subprocess as asubprocess
import subprocess
from typing import Any, AsyncGenerator, Callable, Generator, Generic, TypeVar

from handbrake.canceller import Canceller
from handbrake.errors import CancelledError, HandBrakeError
from handbrake.models.preset import Preset
from handbrake.models.progress import Progress
from handbrake.models.title import TitleSet
from handbrake.models.version import Version

T = TypeVar("T")


class OutputProcessor(Generic[T]):
    """
    Match the beginning and end of an object in command output and
    convert it to a model
    """

    def __init__(
        self,
        start_line: tuple[bytes, bytes],
        end_line: tuple[bytes, bytes],
        converter: Callable[[bytes], T],
    ):
        self.start_line = start_line
        self.end_line = end_line
        self.converter = converter

    def match_start(self, line: bytes) -> bytes | None:
        if line == self.start_line[0]:
            return self.start_line[1]
        return None

    def match_end(self, line: bytes) -> bytes | None:
        if line == self.end_line[0]:
            return self.end_line[1]
        return None

    def convert(self, data: bytes) -> T:
        return self.converter(data)


class CommandRunner:
    def __init__(self, *processors: OutputProcessor):
        self.processors = processors
        self.current_processor: OutputProcessor | None = None
        self.collect: list[bytes] = []

    def process_line(self, line: bytes) -> Any:
        if self.current_processor is None:
            # attempt to start a processor
            for processor in self.processors:
                c = processor.match_start(line)
                if c is not None:
                    self.current_processor = processor
                    self.collect = [c]
                    return
        else:
            # attempt to end the current processor
            c = self.current_processor.match_end(line)
            if c is not None:
                self.collect.append(c)
                res = self.current_processor.convert(b"\n".join(self.collect))
                self.current_processor = None
                self.collect = []
                return res
            # append line to current collect
            self.collect.append(line)

    async def aprocess(
        self,
        cmd: str,
        *args: str,
        cancel: Canceller | None = None,
    ) -> AsyncGenerator[Any, None]:
        aproc = await asubprocess.create_subprocess_exec(
            cmd,
            *args,
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
        )
        try:
            if aproc.stdout is None:
                raise ValueError

            # slurp output while running
            while True:
                if cancel is not None and cancel.is_cancelled():
                    raise CancelledError
                try:
                    # get a whole line; if this returns empty then output has finished
                    line = await asyncio.wait_for(aproc.stdout.readline(), 1)
                    if not line:
                        break
                except asyncio.TimeoutError:
                    pass
                else:
                    o = self.process_line(line.rstrip())
                    if o is not None:
                        yield o

            # raise error on nonzero return code
            if (returncode := await aproc.wait()) != 0:
                raise HandBrakeError(returncode)

        finally:
            # ensure program is terminated on exit
            try:
                aproc.terminate()
            except ProcessLookupError:
                pass

    def process(self, cmd: str, *args: str) -> Generator[Any, None, None]:
        # create process with pipes to output
        proc = subprocess.Popen(
            [cmd, *args],
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
        )
        if proc.stdout is None:
            raise ValueError

        try:
            # slurp stdout line-by-line
            while True:
                stdout = proc.stdout.readline()
                if len(stdout) == 0 and proc.poll() is not None:
                    break
                o = self.process_line(stdout.rstrip())
                if o is not None:
                    yield o

            # slurp the remaining output
            lines = proc.stdout.read()
            for line in lines.splitlines():
                o = self.process_line(line.rstrip())
                if o is not None:
                    yield o

            # raise error on nonzero return code
            if proc.returncode != 0:
                raise HandBrakeError(proc.returncode)
        finally:
            try:
                proc.terminate()
            except ProcessLookupError:
                pass


class VersionCommandRunner(CommandRunner):
    def __init__(self):
        processor = OutputProcessor(
            (b"Version: {", b"{"),
            (b"}", b"}"),
            Version.model_validate_json,
        )
        super().__init__(processor)


class ConvertCommandRunner(CommandRunner):
    def __init__(self):
        processor = OutputProcessor(
            (b"Progress: {", b"{"),
            (b"}", b"}"),
            Progress.model_validate_json,
        )
        super().__init__(processor)


class ScanCommandRunner(CommandRunner):
    def __init__(self):
        progress_processor = OutputProcessor(
            (b"Progress: {", b"{"),
            (b"}", b"}"),
            Progress.model_validate_json,
        )
        titleset_processor = OutputProcessor(
            (b"JSON Title Set: {", b"{"),
            (b"}", b"}"),
            TitleSet.model_validate_json,
        )
        super().__init__(progress_processor, titleset_processor)


class PresetCommandRunner(CommandRunner):
    def __init__(self):
        processor = OutputProcessor(
            (b"{", b"{"),
            (b"}", b"}"),
            Preset.model_validate_json,
        )
        super().__init__(processor)



from handbrake.models.common import HandBrakeModel


class VersionIdentifier(HandBrakeModel):
    major: int
    minor: int
    point: int


class Version(HandBrakeModel):
    arch: str
    name: str
    official: bool
    repo_date: str
    repo_hash: str
    system: str
    type: str
    version: VersionIdentifier
    version_string: str



from pydantic import Field

from handbrake.models.common import HandBrakeModel


class ProgressScanning(HandBrakeModel):
    preview: int
    preview_count: int
    progress: float
    sequence_id: int = Field(alias="SequenceID")
    title: int
    title_count: int


class ProgressWorking(HandBrakeModel):
    eta_seconds: int = Field(alias="ETASeconds")
    hours: int
    minutes: int
    pass_: int = Field(alias="Pass")
    pass_count: int
    pass_id: int = Field(alias="PassID")
    paused: int
    progress: float
    rate: float
    rate_avg: float
    seconds: int
    sequence_id: int = Field(alias="SequenceID")


class ProgressWorkDone(HandBrakeModel):
    error: int
    sequence_id: int = Field(alias="SequenceID")


class Progress(HandBrakeModel):
    scanning: ProgressScanning | None = None
    working: ProgressWorking | None = None
    work_done: ProgressWorkDone | None = None
    state: str

    @property
    def percent(self) -> float:
        if self.scanning is not None:
            return self.scanning.progress * 100
        elif self.working is not None:
            return self.working.progress * 100
        elif self.work_done is not None:
            return 100
        return 0

    @property
    def task_description(self) -> str:
        if self.state == "SCANNING":
            return "scanning title sets"
        elif self.state == "WORKING":
            return "processing title"
        elif self.state == "WORKDONE":
            return "done"
        return ""



import os
from typing import Iterable, Literal, TypedDict

from handbrake.models.common import Offset

AudioSelection = Literal["all", "first", "none"]
SubtitleSelection = Literal["all", "first", "scan", "none"]


class ConvertOpts(TypedDict, total=False):
    chapters: int | tuple[int, int]
    angle: int
    previews: tuple[int, bool]
    start_at_preview: int
    start_at: Offset
    stop_at: Offset
    audio: int | Iterable[int] | AudioSelection
    subtitles: int | Iterable[int] | SubtitleSelection
    preset: str
    preset_files: Iterable[str | os.PathLike]
    preset_from_gui: bool
    no_dvdnav: bool


def generate_convert_args(
    input: str | os.PathLike,
    output: str | os.PathLike,
    title: int | Literal["main"],
    opts: ConvertOpts | None,
) -> list[str]:
    if title == 0:
        raise ValueError("invalid title")

    # generate base args
    args: list[str] = [
        "--json",
        "-i",
        str(input),
        "-o",
        str(output),
    ]
    if title == "main":
        args += ["--main-feature"]
    else:
        args += ["-t", str(title)]

    # generate opts
    if opts is not None:
        # generate list of preset import files
        preset_import_files: list[str] = []
        if preset_files := opts.get("preset_files"):
            preset_import_files += [str(f) for f in preset_files]

        # preset args
        if len(preset_import_files) > 0:
            args += [
                "--preset-import-file",
                " ".join(preset_import_files),
            ]
        if preset := opts.get("preset"):
            args += ["--preset", preset]
        if opts.get("preset_from_gui"):
            args += ["--preset-import-gui"]

        # dvdnav arg
        if opts.get("no_dvdnav"):
            args += ["--no-dvdnav"]

        # chapters arg
        if (chapters := opts.get("chapters")) is not None:
            if isinstance(chapters, tuple):
                args += ["-c", f"{chapters[0]}-{chapters[1]}"]
            elif isinstance(chapters, int):
                args += ["-c", str(chapters)]

        # angle arg
        if (angle := opts.get("angle")) is not None:
            args += ["--angle", str(angle)]

        # preview args
        if previews := opts.get("previews"):
            args += [
                "--previews",
                f"{previews[0]}:{int(previews[1])}",
            ]
        if opts.get("start_at_preview") is not None:
            args += ["--start-at-preview", str(opts.get("start_at_preview"))]

        # start/stop args
        if start_at := opts.get("start_at"):
            args += [
                "--start-at",
                f"{start_at.unit}:{start_at.count}",
            ]
        if stop_at := opts.get("stop_at"):
            args += [
                "--stop-at",
                f"{stop_at.unit}:{stop_at.count}",
            ]

        # audio args
        if (audio := opts.get("audio")) is not None:
            if isinstance(audio, int):
                args += ["--audio", str(audio)]
            elif audio == "all":
                args += ["--all-audio"]
            elif audio == "first":
                args += ["--first-audio"]
            elif audio == "none":
                args += ["--audio", "none"]
            else:
                args += ["--audio", ",".join(str(a) for a in audio)]

        # subtitle args
        if (subtitles := opts.get("subtitles")) is not None:
            if isinstance(subtitles, int):
                args += ["--subtitle", str(subtitles)]
            elif subtitles == "all":
                args += ["--all-subtitles"]
            elif subtitles == "first":
                args += ["--first-subtitle"]
            elif subtitles == "none":
                args += ["--subtitle", "none"]
            elif subtitles == "scan":
                args += ["--subtitle", "scan"]
            else:
                args += ["--subtitle", ",".join(str(s) for s in subtitles)]

    return args


def generate_scan_args(
    input: str | os.PathLike,
    title: int | Literal["main", "all"],
) -> list[str]:
    args: list[str] = ["--json", "-i", str(input), "--scan"]
    if title == "main":
        args += ["--main-feature"]
    elif title == "all":
        args += ["-t", "0"]
    else:
        args += ["-t", str(title)]

    return args



import asyncio
import json
import os
from dataclasses import dataclass
from datetime import timedelta
from io import TextIOBase
from os import PathLike
from time import sleep
from typing import Iterable, Literal

from handbrake import HandBrake
from handbrake.canceller import Canceller
from handbrake.models.common import Duration, Fraction
from handbrake.models.preset import Preset, PresetGroup
from handbrake.models.progress import (
    Progress,
    ProgressScanning,
    ProgressWorkDone,
    ProgressWorking,
)
from handbrake.models.title import Color, Geometry, Title, TitleSet
from handbrake.models.version import Version, VersionIdentifier
from handbrake.opts import ConvertOpts
from handbrake.progresshandler import ProgressHandler


class DictJSONEncoder(json.JSONEncoder):
    def default(self, o: object):
        return o.__dict__


@dataclass
class MockTitle:
    index: int
    runtime: timedelta

    def get_title(self) -> Title:
        return Title(
            angle_count=1,
            audio_list=[],
            chapter_list=[],
            color=Color(
                bit_depth=-1,
                chroma_location=-1,
                chroma_subsampling="",
                format=1,
                matrix=1,
                primary=1,
                range=1,
                transfer=1,
            ),
            crop=(0, 0, 0, 0),
            duration=Duration.from_timedelta(self.runtime),
            frame_rate=Fraction(num=30, den=1),
            geometry=Geometry(height=360, width=480, PAR=Fraction(num=480, den=360)),
            index=self.index,
            interlace_detected=False,
            loose_crop=(-1, -1, -1, -1),
            metadata={},
            name=f"Title {self.index}",
            path=f"mock/{self.index}",
            playlist=-1,
            subtitle_list=[],
            type=-1,
            video_codec="AV1",
        )


class MockHandBrake(HandBrake):
    def __init__(
        self,
        title_runtime_minutes: Iterable[int],
        touch: bool = False,
        scan_factor: float = 0.0003,
        convert_factor: float = 0.001,
    ):
        self.scan_factor = scan_factor
        self.convert_factor = convert_factor
        self.titles = [
            MockTitle(i, timedelta(minutes=m))
            for i, m in enumerate(title_runtime_minutes, 1)
        ]
        self.touch = touch
        self.main_title = max(
            range(len(self.titles)),
            key=lambda i: self.titles[i].runtime,
        )

    def version(self) -> Version:
        return Version(
            arch="Python",
            name="HandBrake (mock)",
            official=False,
            repo_date="",
            repo_hash="",
            system="Python",
            type="mock",
            version=VersionIdentifier(major=0, minor=0, point=0),
            version_string="0.0.0",
        )

    async def version_async(self, cancel: Canceller | None = None) -> Version:
        _ = cancel
        return Version(
            arch="Python",
            name="HandBrake (mock)",
            official=False,
            repo_date="",
            repo_hash="",
            system="Python",
            type="mock",
            version=VersionIdentifier(major=0, minor=0, point=0),
            version_string="0.0.0",
        )

    def convert_title(
        self,
        input: str | os.PathLike,
        output: str | os.PathLike,
        title: int | Literal["main"],
        opts: ConvertOpts | None = None,
        progress_handler: ProgressHandler | None = None,
    ):
        if title == "main":
            t = self.titles[self.main_title]
        else:
            t = self.titles[title - 1]
        total = int(t.runtime.total_seconds())
        if self.touch:
            with open(output, "w") as f:
                d = {
                    "input": input,
                    "title": title,
                    **(opts or {}),
                }
                json.dump(d, f)
        for i in range(total):
            sleep(self.convert_factor)
            if progress_handler is not None:
                pw = ProgressWorking(
                    ETASeconds=int(self.convert_factor * (total - i)),
                    hours=0,
                    minutes=i,
                    Pass=1,
                    pass_count=1,
                    PassID=1,
                    paused=0,
                    progress=i / total,
                    rate=1,
                    rate_avg=1,
                    seconds=0,
                    SequenceID=0,
                )
                progress_handler(Progress(working=pw, state="WORKING"))
        if progress_handler is not None:
            pd = ProgressWorkDone(error=0, SequenceID=0)
            progress_handler(Progress(work_done=pd, state="WORKDONE"))

    async def convert_title_async(
        self,
        input: str | os.PathLike,
        output: str | os.PathLike,
        title: int | Literal["main"],
        opts: ConvertOpts | None = None,
        progress_handler: ProgressHandler | None = None,
        cancel: Canceller | None = None,
    ):
        if title == "main":
            t = self.titles[self.main_title]
        else:
            t = self.titles[title - 1]
        total = int(t.runtime.total_seconds())
        if self.touch:
            with open(output, "w") as f:
                d = {
                    "input": input,
                    "title": title,
                    **(opts or {}),
                }
                json.dump(d, f)
        for i in range(total):
            await asyncio.sleep(self.convert_factor)
            if cancel and cancel.is_cancelled():
                return
            if progress_handler is not None:
                pw = ProgressWorking(
                    ETASeconds=int(self.convert_factor * (total - i)),
                    hours=0,
                    minutes=i,
                    Pass=1,
                    pass_count=1,
                    PassID=1,
                    paused=0,
                    progress=i / total,
                    rate=1,
                    rate_avg=1,
                    seconds=0,
                    SequenceID=0,
                )
                progress_handler(Progress(working=pw, state="WORKING"))
        if progress_handler is not None:
            pd = ProgressWorkDone(error=0, SequenceID=0)
            progress_handler(Progress(work_done=pd, state="WORKDONE"))

    def scan_titles(
        self,
        input: str | PathLike,
        title: int | Literal["main", "all"],
        progress_handler: ProgressHandler | None = None,
    ) -> TitleSet:
        _ = input
        if title == 0 or title == "all":
            main_feature = self.main_title + 1
            titles = [t for t in self.titles]
        elif title == "main":
            main_feature = self.main_title + 1
            titles = [self.titles[self.main_title]]
        else:
            main_feature = title + 1
            titles = [self.titles[title - 1]]

        partial = 0
        overall_total = sum(int(t.runtime.total_seconds()) for t in titles)
        for i, t in enumerate(titles):
            total = int(t.runtime.total_seconds())
            for p in range(total):
                sleep(self.scan_factor)
                if progress_handler is not None:
                    ps = ProgressScanning(
                        preview=0,
                        preview_count=0,
                        progress=(partial + p) / overall_total,
                        SequenceID=0,
                        title=i + 1,
                        title_count=len(self.titles),
                    )
                    progress_handler(Progress(scanning=ps, state="SCANNING"))
            partial += total

        return TitleSet(
            main_feature=main_feature,
            title_list=[t.get_title() for t in titles],
        )

    async def scan_titles_async(
        self,
        input: str | PathLike,
        title: int | Literal["main", "all"],
        progress_handler: ProgressHandler | None = None,
        cancel: Canceller | None = None,
    ) -> TitleSet:
        _ = input
        if title == 0 or title == "all":
            main_feature = self.main_title + 1
            titles = [t for t in self.titles]
        elif title == "main":
            main_feature = self.main_title + 1
            titles = [self.titles[self.main_title]]
        else:
            main_feature = title + 1
            titles = [self.titles[title - 1]]

        partial = 0
        overall_total = sum(int(t.runtime.total_seconds()) for t in titles)
        for i, t in enumerate(titles):
            total = int(t.runtime.total_seconds())
            for p in range(total):
                await asyncio.sleep(self.scan_factor)
                if cancel and cancel.is_cancelled():
                    return TitleSet(main_feature=0, title_list=[])
                if progress_handler is not None:
                    ps = ProgressScanning(
                        preview=0,
                        preview_count=0,
                        progress=(partial + p) / overall_total,
                        SequenceID=0,
                        title=i + 1,
                        title_count=len(self.titles),
                    )
                    progress_handler(Progress(scanning=ps, state="SCANNING"))
            partial += total

        return TitleSet(
            main_feature=main_feature,
            title_list=[t.get_title() for t in titles],
        )

    def get_preset(self, name: str) -> Preset:
        _ = name
        return Preset(version_major=0, version_minor=0, version_micro=0, preset_list=[])

    def list_presets(self) -> list[PresetGroup]:
        return []

    def load_preset_from_file(self, file: str | PathLike | TextIOBase) -> Preset:
        _ = file
        return Preset(version_major=0, version_minor=0, version_micro=0, preset_list=[])

    def save_preset_to_file(self, file: str | PathLike | TextIOBase, preset: Preset):
        _ = file
        _ = preset
